import { ArkErrors, intrinsic, rootNode } from "@ark/schema";
import { wellFormedIntegerMatcher, wellFormedNumberMatcher } from "@ark/util";
import { scope } from "../scope.js";
import { arkNumber } from "./number.js";
import { creditCardMatcher, isLuhnValid } from "./utils/creditCard.js";
import { iso8601Matcher } from "./utils/date.js";
import { arkIp } from "./utils/ip.js";
import { regexStringNode } from "./utils/regex.js";
import { arkUuid } from "./utils/uuid.js";
// Non-trivial expressions should have an explanation or attribution
const isValidUrl = (s) => {
    if (URL.canParse)
        return URL.canParse(s);
    // Can be removed once Node 18 is EOL
    try {
        new URL(s);
    }
    catch {
        return false;
    }
    return true;
};
const url = rootNode({
    domain: "string",
    predicate: {
        meta: "a URL",
        predicate: isValidUrl
    }
});
// https://www.regular-expressions.info/email.html
const emailMatcher = /^[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Za-z]{2,}$/;
const email = regexStringNode(emailMatcher, "an email address");
// https://semver.org/
const semverMatcher = /^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$/;
const semver = regexStringNode(semverMatcher, "a semantic version (see https://semver.org/)");
const creditCard = rootNode({
    domain: "string",
    pattern: {
        meta: "a credit card number",
        rule: creditCardMatcher.source
    },
    predicate: {
        meta: "a credit card number",
        predicate: isLuhnValid
    }
});
const numericString = regexStringNode(wellFormedNumberMatcher, "a well-formed numeric string");
const integerString = regexStringNode(wellFormedIntegerMatcher, "a well-formed integer string");
const epoch = integerString
    .narrow((s, ctx) => {
    // we know this is safe since it has already
    // been validated as an integer string
    const n = Number.parseInt(s);
    const out = arkNumber.submodule.epoch(n);
    if (out instanceof ArkErrors) {
        ctx.errors.merge(out);
        return false;
    }
    return true;
})
    .describe("an integer string representing a safe Unix timestamp");
const iso8601 = regexStringNode(iso8601Matcher, "an ISO 8601 (YYYY-MM-DDTHH:mm:ss.sssZ) date");
const keywords = scope({
    creditCard,
    email,
    uuid: arkUuid.submodule,
    url,
    semver,
    ip: arkIp.submodule
}, { prereducedAliases: true }).export();
const submodule = scope({
    $root: intrinsic.string,
    numeric: numericString,
    integer: integerString,
    alpha: regexStringNode(/^[A-Za-z]*$/, "only letters"),
    alphanumeric: regexStringNode(/^[A-Za-z\d]*$/, "only letters and digits 0-9"),
    digits: regexStringNode(/^\d*$/, "only digits 0-9"),
    lower: regexStringNode(/^[a-z]*$/, "only lowercase letters"),
    upper: regexStringNode(/^[A-Z]*$/, "only uppercase letters"),
    iso8601,
    epoch,
    ...keywords
}, {
    prereducedAliases: true
}).export();
export const arkString = {
    keywords,
    submodule
};
